#include "CcProcessor.h"

CcProcessor::CcProcessor(CcExtractor *x)
	:cThread("CC_OSD")
{
	dprint(1, "CcProcessor::CcProcessor()");
	if(x==NULL)
		dprint(1, "CcProcessor::CcProcessor: NO VALID EXTRACTOR WAS SPECIFIED");
	osd = NULL;
	extractor = x;
	type = config.getCcType();
	
	vsPos = 0;
	currentContext = &realContext;
	InitContext(currentContext);
	
	font = cFont::GetFont(fontFix);

	osdInitialized = false;
	
	////
	nRollup = 0;
  bTextMode = false;
  
  crtContext = new sCcContext();
  crtContext->nBufferPos = 0;  
  crtContext->nX = 0;
  crtContext->nY = 0;
  
  bufferContext = new sCcContext();
  bufferContext->nBufferPos = 0; 
  bufferContext->nX = 0;
  bufferContext->nY = 0;    
	////
}

void CcProcessor::InitContext(CcText *cx){
	cx->posX		= 0;
	cx->posY		= 0;
	cx->length	= 0;
}

CcProcessor::~CcProcessor()
{
	Stop();
	if(osd!=NULL){
		delete osd;
		osd = NULL;
	}
	
	////resetBuffer();	
	delete bufferContext;
	delete crtContext;
}

void CcProcessor::PrintBuffer(CcText **buffer, int size){

	for(int i=0; i<size; i++)
		Print(buffer[i], ((i+1)==size));
}

void CcProcessor::InitOsd(){
	tArea Areas[] = { { 0, 0, OSDWIDTH, OSDHEIGHT, 2 } };
	if(osd==NULL){	
		osd = cOsdProvider::NewOsd(START_X, START_Y, true);
		
		if (osd->CanHandleAreas(Areas, sizeof(Areas) / sizeof(tArea)) == oeOk){		  
		  osd->SetAreas(Areas, sizeof(Areas) / sizeof(tArea));
		  
		  dprint(1, "OSD was initialized by one slice (W:%d x H: %d)", OSDWIDTH, OSDHEIGHT);
		}
		else {
		  tArea Areas[] = { { 0, 0, 				(OSDWIDTH-1), OSDHEIGHT/3,		2 },
		                    { 0, OSDHEIGHT/3+1,	(OSDWIDTH-1), 2*OSDHEIGHT/3,	2 },
		                    { 0, 2*OSDHEIGHT/3+1, (OSDWIDTH-1), OSDHEIGHT,		2 }
		                 };
		  osd->SetAreas(Areas, sizeof(Areas) / sizeof(tArea));
		  
		  dprint(1, "OSD was initialized by thirds (W:%d x H: %d)", OSDWIDTH, OSDHEIGHT);	    
		}	
		
		osdInitialized = true;
	  dprint(1, "OSD was set.");
	  
	  ClearScreen();
	}
	else
		printf("CC: OSD Not Supported!");
}

void CcProcessor::Print(CcText *data, bool flush){
	cMutexLock lock(&osdMutex);
	if(!osdInitialized)
				InitOsd();
				
	if(osd == NULL){
		dprint(1, "CcProcessor::PrintBuffer - osd does not exist");
		return;
	}
				
	osd->DrawText(
		data->posX * CHAR_WIDTH,
		data->posY * CHAR_HEIGHT,
		(const char*)data->text,
		clrWhite, clrBlack, font
	 );
	 if(flush)
	 	osd->Flush();
}

void CcProcessor::ClearScreen(void){	
	cMutexLock lock(&osdMutex);
	
	if(!osdInitialized)
				InitOsd();
				
	if(osd){
		dprint(1, "CcProcessor::ClearScreen - [Screen cleared]");
		osd->DrawRectangle(0, 0, OSDWIDTH, OSDHEIGHT, clrTransparent);
		osd->Flush();    
	}
}

void CcProcessor::Stop(){
	dprint(1, "CcProcessor::Stop()");
	Cancel(5);
}

int CcProcessor::UndoubleCommands(byte* commands, int size, byte *buffer){
	int c = 0;
	for(int i=0; i<size; i+=2){
		buffer[c++] = commands[i];
		buffer[c++] = commands[i+1];
	
		if(
				(size >= 4) &&
				(commands[i]	== commands[i+2]) &&
				(commands[i+1]	== commands[i+3])
			)
			i+=2;
	}
	return c;
}
/*
bool CcProcessor::Process(CC_Data *incoming){
		
	if(data->type==CC_TEXT){
		//removing the unwanted 0x80
		while((incoming->length > 0) && (incoming->data[incoming->length-1]==(byte)0x80))
			incoming->length--;
			
		char *text = new char[incoming->length+1];
		memcpy(text, incoming->data, incoming->length * sizeof(byte));
		text[data->length] = 0x00;
		printf("%s\n", text);
		delete text;		
	}
	else
		if(data->type==CC_INFO){
			byte *data = new byte[incoming->length];
			int length = UndoubleCommands(incoming->data, incoming->length, data);
			
			for(int i=0; i<length; i+=2){
				DoCommand(bWriteInBuffer ? bufferContext : crtContext, data->data[i], data->data[i+1]);	
			}
			
			delete data;
		}
	
	return true;
}*/

//! returns false if the data should be preserved
bool CcProcessor::Process(CC_Data* data, bool bWorkInBuffer){
	//dprint(1, "CC:Processing CC...");
	bool bOK = false;
	byte last_hi = 0xFF, last_lo = 0xFF;
	if(data==NULL){
		dprint(2, "CC:No CC data...");
		return bOK;
	}
	
	bOK = true;
	
	//we cannot write in buffer while we are processing the buffer
	if(bWorkInBuffer)		
		bWriteInBuffer = false;	
		
	if(data->type==0x01){ //data
		for(int i=0; i<data->length; i+=2){
			dprint(3, "\nbWorkInBuffer=%s", (bWorkInBuffer ? "=== TRUE ===" : "FALSE"));
			if((data->data[i]==last_hi)	&& (data->data[i+1]==last_lo))
				continue;
			DoCommand(bWriteInBuffer ? bufferContext : crtContext, data->data[i], data->data[i+1]);	
			if(bWriteInBuffer && !bWorkInBuffer){
				if(!bSkipBuffer){
					WriteInBuffer(data,i,2);
					dprint(3, "COMMAND was written in buffer");
				}				
			}
			else
				bWriteInBuffer = false;
			last_hi = data->data[i];
			last_lo = data->data[i+1];
		}
	}
	else{		
		if(bWriteInBuffer && (bufferContext->nBufferPos < CC_BUFFER_SIZE)){
			WriteInBuffer(data,0,data->length);
			dprint(1, "Write TEXT in buffer");
		}
		else
			
			if(nRollup>0){
				dprint(1, "Roll UP mode #%d detected.", nRollup);				
				scroller.AddText((char *)data->data, data->length);
				char **lines = scroller.GetLines();
				//int all = scroller.CountLines();
				dprint(1, "RU%d lines: %d", nRollup, scroller.CountLines());
				PrintRollUp(crtContext, lines, scroller.GetHeight());
			}
			else				
				PrintLine(crtContext, (char *) data->data, data->length);				
	}
	
	return bOK;
}


void CcProcessor::Action(){
	dprint(1, "CcProcessor::Action()");
	int t0 = 0;
	while(Running()){
		t0 = time_ms();
		
		if(extractor==NULL){
			dprint(1,"CcProcessor::Action() NO EXTRACTOR!...");
			Cancel();
			break;
		}
		
		CC_Data *cc = extractor->PopCC();
		if(cc!=NULL){
			dprint(3,"CcProcessor::Action() Processing in Action...");
			if(Process(cc, false)){
					//if the cc data was not stored in the cc bufer then they must be erased
						if(cc->data!=NULL)
							delete [] cc->data;
					delete cc;
				}
		}
		else
			dprint(3,"CcProcessor::Action() No CC Data!");
		
		cCondWait::SleepMs(OSD_FREQ - (time_ms() - t0));
	}
	
	dprint(1, "CcProcessor::Action() FINISHED");
}
/*
bool CcProcessor::DoCommand(CcText* info, byte hi, byte lo){
	dprint(3, "Do Command...");
		
	//Global commands
	byte channels4g[] = {0x94, 0x1c, 0x15, 0x9d};
		
	//Processing general commands
	if(channels4g[type & 0x03]==hi){
				
		switch(lo){
			
			//RCL - Resume Caption Loading
			case 0x20:
				dprint(2, "{RCL}");					
				info->style = POPUP;
				break;
				
			//RDC - Resume Direct Captioning
			case 0x29:
				dprint(2, "{RDC}");
				info->style = PAINTON;
				break;
				
			//RU2 - 	Roll-Up Captions--2 Rows
			case 0x25:
				dprint(1, "{RU2}");
				info->style = ROLLUP2;
				break;
				
			//RU3 - 	Roll-Up Captions--3 Rows
			case 0x26:
				dprint(1, "{RU3}");
				info->style = ROLLUP3;
				break;
				
			//RU4 - 	Roll-Up Captions--4 Rows
			case 0xA7:
				dprint(1, "{RU4}");
				info->style = ROLLUP4;
				break;
				
			//TR - Text Restart
			case 0x2A:
				dprint(1, "{TR}");
				info->posX = 0;
				info->posY = 0;
				info->style = TEXT;
				break;
				
			//RTD - Resume Text Display	
			case 0xAB:
				dprint(1, "{RTD}");
				info->style = TEXT;
				break;
				
			//EDM - Erase Displayed Memory
			case 0x2C:				
				dprint(1, "{EDM}");
				ClearScreen();
				break;
				
			//ENM - Erase Non-displayed Memory
			case 0xAE:
				dprint(2, "{ENM}");
				//ResetBuffer();
				break;
				
			//EOC - End Of Caption
			case 0x2F:
				dprint(1, "{EOC}");
				ClearScreen();
				resetBuffer();		
				break;
////				
			//backspace
			case 0xA1:
				dprint(1, "{BS}");
				scroller.Backspace();			
			  bDone = true;
				break;
				
			//delete to end of row
			case 0xA4:
				dprint(1, "{DER}");
				//clear current line starting from current char
				ClearLine(context->nY, context->nX, -1);
				bDone = true;
				break;
				
			//carriage return
			case 0xAD:
				dprint(1, "{CR}");
				scroller.Scroll();					
				bDone = true;
				break;
////
			default:
				return CheckPAC(info, hi, lo);			
		}	
	
		
		if(lo!=0xA1 && lo!=0xA4 && lo!=0xAD && lo!=0x25 && lo!=0x26 && lo!=0xA7){
			//scroller.Clear();
			info->style = UNDEFINED;
		}
			return true;		
	}
	
	return false;
}
*/

bool CcProcessor::DoCommand(sCcContext *context, byte hi, byte lo){
	dprint(3, "Do Command...%0X-%0X", hi, lo);
	if(hi==lo && hi==0x80){
		cCondWait::SleepMs(67);
		return false;
	}
	
	byte pac_lines[] = {0x91/*0*/, 0x91/*1*/, 0x92/*2*/, 0x92/*3*/, 0x15/*4*/, 0x15/*5*/, 0x16/*6*/, 0x16/*7*/, 
		0x97/*8*/, 0x97/*9*/, 0x10/*10*/, 0x13/*11*/, 0x13/*12*/, 0x94/*13*/, 0x94/*14*/};
	byte pac_cols[] = {0x40, 0xC1, 0xC2, 0x43, 0xC4, 0x45, 0x46, 0xC7, 0xC8, 0x49, 0x4A, 0xCB, 0x4C, 0xCD, 0xCE, 0x4F, 0xD0, 0x51, 0x52, 0xD3, 0x54, 0xD5, 0xD6, 0x57, 0x58, 0xD9, 0xDA, 0x5B, 0xDC, 0x5D, 0x5E, 0xDF,
										 0xE0, 0x61, 0x62, 0xE3, 0x64, 0xE5, 0xE6, 0x67, 0x68, 0xE9, 0xEA, 0x6B, 0xEC, 0x6D, 0x6E, 0xEF, 0x70, 0xF1, 0xF2, 0x73, 0xF4, 0x75, 0x76, 0xF7, 0xF8, 0x79, 0x7A, 0xFB, 0xFC, 0xFD, 0xFE, 0x7F};
	bool bIsPAC = false;
	
	//Global commands
	byte channels4g[] = {0x94, 0x1c, 0x15, 0x9d};
	
	//Tab positionning commands
	byte channels2p[] = {0x97, 0x1f};
	
	bool bDone = false;
	bSkipBuffer = false;
	
	//Processing general commands
	if(channels4g[type]==hi){
		bTextMode = false;
		
		switch(lo){
			
			//RCL - Resume Caption Loading
			case 0x20:
				dprint(2, "{RCL}");					
				bWriteInBuffer = true;
				bSkipBuffer = true;
				bDone=true;
				break;
				
			//RDC - Resume Direct Captioning
			case 0x29:
				dprint(2, "{RDC}");
				bWriteInBuffer = false;
				bDone = true;
				break;
				
			//RU2 - 	Roll-Up Captions--2 Rows
			case 0x25:
				dprint(1, "{RU2}");
				SetRollUp(context, 2);
				bDone = true;
				break;
				
			//RU3 - 	Roll-Up Captions--3 Rows
			case 0x26:
				dprint(1, "{RU3}");
				SetRollUp(context, 3);
				bDone = true;
				break;
				
			//RU4 - 	Roll-Up Captions--4 Rows
			case 0xA7:
				dprint(1, "{RU4}");
				SetRollUp(context, 4);
				bDone = true;
				break;
				
			//TR - Text Restart
			case 0x2A:
				dprint(1, "{TR}");
				context->nX = 0;
				context->nY = 0;
				bTextMode = true;
				bDone = true;
				break;
				
			//RTD - Resume Text Display	
			case 0xAB:
				dprint(1, "{RTD}");
				bTextMode = true;
				bDone = true;
				break;
				
			//EDM - Erase Displayed Memory
			case 0x2C:				
				dprint(1, "{EDM}");
				ClearScreen();
				bSkipBuffer = true;
				bDone = true;
				break;
				
			//ENM - Erase Non-displayed Memory
			case 0xAE:
				dprint(2, "{ENM}");
				bSkipBuffer = true;
				ClearBuffer();				
				bDone = true;
				break;
				
			//EOC - End Of Caption
			case 0x2F:
				dprint(1, "{EOC}");
				ClearScreen();
				if(bWriteInBuffer){					
					bSkipBuffer = true;
					dprint(3, "Found %d commands in buffer.", context->nBufferPos);
					for(int i=0; i<context->nBufferPos; i++){
						bWriteInBuffer = false;
						Process(&context->commands[i], true);								
					}												
				}
				ClearBuffer();
				
				bWriteInBuffer = false;				
				bSkipBuffer = true;
				bDone = true;				
				break;
				
			//backspace
			case 0xA1:
				dprint(1, "{BS}");
				scroller.Backspace();			
			  bDone = true;
				break;
				
			//delete to end of row
			case 0xA4:
				dprint(1, "{DER}");
				//clear current line starting from current char
				ClearLine(context->nY, context->nX, -1);
				bDone = true;
				break;
				
			//carriage return
			case 0xAD:
				dprint(1, "{CR}");
				scroller.Scroll();					
				bDone = true;
				break;
		}
		
		if(bDone){
			if(lo!=0xA1 && lo!=0xA4 && lo!=0xAD && lo!=0x25 && lo!=0x26 && lo!=0xA7){
				scroller.Clear();
				nRollup = 0;
			}
			return true;
		}
	}
	//TO1, TO2, TO3
	if(
			(
			(type<2 && channels2p[type/2]==hi) ||
			(type>=2 && channels2p[type/2]==hi)
			) &&
			(lo==0xA1 || lo==0xA2 || lo==0x23)
		){	
			context->nX += (lo & 0x0F);	
			dprint(2, "{TO%d}", lo & 0x0F);		
			return true;
	}	
	
	int pacY = 0;
	//Processing PAC
	for(int i=0; i<15; i++)
		if(type%2==0) // CC1, CC3
			if(pac_lines[i]==hi){
				bIsPAC = true;
				pacY = i;
				break;
			}				
		else //CC2, CC4
			if((pac_lines[i]-0x78	)==hi){
				bIsPAC = true;
				pacY = i;
				break;
			}
			
			
	if(bIsPAC && (nRollup==0)){		
		dprint(1, "{PAC}");
		context->bUnderline = ((lo & 0x01) == 	0x01);
		for(int i=0; i<64; i++)//i=19
			if(pac_cols[i]==lo){
				int p = i%32;
				pacY += (i<32) ? 0 : 1;
				if(p<18)
					context->nX = 0;
				else
					context->nX = 4 *(8 - (32-p)/2);
				context->nY = pacY;
				
				dprint(1,"{%2X%2X}=[%d %d]", hi, lo, context->nX, context->nY);	
				return true;
			}
	}
	

	dprint(2, "{!%2X%2X!}", hi, lo);
	return false;
}	
bool CcProcessor::CheckPAC(CcText *info, byte hi, byte lo){	
	byte pac_lines[] = {0x91/*0*/, 0x91/*1*/, 0x92/*2*/, 0x92/*3*/, 0x15/*4*/, 0x15/*5*/, 0x16/*6*/, 0x16/*7*/, 
		0x97/*8*/, 0x97/*9*/, 0x10/*10*/, 0x13/*11*/, 0x13/*12*/, 0x94/*13*/, 0x94/*14*/};
	byte pac_cols[] = {0x40, 0xC1, 0xC2, 0x43, 0xC4, 0x45, 0x46, 0xC7, 0xC8, 0x49, 0x4A, 0xCB, 0x4C, 0xCD, 0xCE, 0x4F, 0xD0, 0x51, 0x52, 0xD3, 0x54, 0xD5, 0xD6, 0x57, 0x58, 0xD9, 0xDA, 0x5B, 0xDC, 0x5D, 0x5E, 0xDF,
										 0xE0, 0x61, 0x62, 0xE3, 0x64, 0xE5, 0xE6, 0x67, 0x68, 0xE9, 0xEA, 0x6B, 0xEC, 0x6D, 0x6E, 0xEF, 0x70, 0xF1, 0xF2, 0x73, 0xF4, 0x75, 0x76, 0xF7, 0xF8, 0x79, 0x7A, 0xFB, 0xFC, 0xFD, 0xFE, 0x7F};
	//Tab positionning commands
	byte channels2p[] = {0x97, 0x1f};
	bool bIsPAC = false;
	
	//TO1, TO2, TO3
	if(
			(
				((type & 0x03)<2 && channels2p[(type & 0x03)/2]==hi) ||
				((type & 0x03)>=2 && channels2p[(type & 0x03)/2]==hi)
			) &&
				(lo==0xA1 || lo==0xA2 || lo==0x23)
		){	
			info->posX += (lo & 0x0F);	
			dprint(2, "{TO%d}", lo & 0x0F);		
			return true;
	}	
	
	int pacY = 0;
	//Processing PAC
	for(int i=0; i<15; i++)
		if(type%2==0) // CC1, CC3
			if(pac_lines[i]==hi){
				bIsPAC = true;
				pacY = i;
				break;
			}				
		else //CC2, CC4
			if((pac_lines[i]-0x78	)==hi){
				bIsPAC = true;
				pacY = i;
				break;
			}
			
			
	if(bIsPAC && NOT_ROLLUP(info->style)){		
		dprint(1, "{PAC}");
		info->underline = ((lo & 0x01) == 	0x01);
		for(int i=0; i<64; i++)//i=19
			if(pac_cols[i]==lo){
				int p = i%32;
				pacY += (i<32) ? 0 : 1;
				if(p<18)
					info->posX = 0;
				else
					info->posX = 4 *(8 - (32-p)/2);
				info->posY = pacY;
				
				dprint(1,"{%2X%2X}=[%d %d]", hi, lo, info->posX, info->posY);	
				return true;
			}
	}
	

	dprint(2, "{!%2X%2X!}", hi, lo);
	return false;
}

bool CcProcessor::WriteInBuffer(CC_Data *data, int pos, int len){
	bool bOk = false;

	if(bufferContext->nBufferPos <CC_BUFFER_SIZE){
		bufferContext->commands[bufferContext->nBufferPos].type		= data->type;
		bufferContext->commands[bufferContext->nBufferPos].length	= len;
		bufferContext->commands[bufferContext->nBufferPos].data		= new byte[len];
		
		int i = pos;
		int j = 0;
		
		for(i = pos; i<pos+len; i++)
			if((data->data[i]>=0x20) && (data->data[i]<0x80))
				bufferContext->commands[bufferContext->nBufferPos].data[j++] = data->data[i];
				
		bufferContext->commands[bufferContext->nBufferPos].data[j] = 0x00;
		bufferContext->nBufferPos++;
		bOk = true;
		
		
	}
	return bOk;
}

bool CcProcessor::PrintLine(sCcContext *cx, char *row, int len){
	dprint(3, "Write TEXT line on screen (%d chars), row is %s", len, (row==NULL) ? "NULL" : "OK");
		
	if(osd==NULL){
		dprint(3, "CC: Osd was not previously created.");
		return false;
	}
	
	if(row==NULL){
		dprint(3, "CC Error: Cannot print a null line.");
		return false;
	}
		
	cMutexLock lock(&osdMutex);
	char line[CC_WIDTH];
	
	int c=0;	
	for(c=0; (c<CC_WIDTH) && (c<len); c++)	
		line[c] = row[c];
	line[c] = 0x00;	
	
	if(c>0 ){
		dprint(1, "DISPLAYED TEXT [%d|%d,%d|%d]: %s", cx->nX , CHAR_WIDTH, cx->nY, CHAR_HEIGHT, line);		
		osd->DrawText(cx->nX * CHAR_WIDTH, cx->nY * CHAR_HEIGHT, line, clrWhite, clrBlack, font);  	
		osd->Flush();
		cx->nX += len;		
  }
	  	
	return true;
}

bool CcProcessor::PrintRollUp(sCcContext *cx, char **text, int lines){
	cMutexLock lock(&osdMutex);
	if(osd!=NULL){
		ClearScreen();
		
		dprint(3, "Printing RU%d (%d lines)", nRollup, lines);
		
		for(int i=0; i<lines; i++){		
			dprint(3, "ROLLUP TEXT [%d|%d,%d|%d]: %s", cx->nX, cx->nY+i , CHAR_WIDTH, CHAR_HEIGHT, text[i]==NULL ? "NULL" : text[i]);
			if(text[i]!=NULL)			
				osd->DrawText(cx->nX * CHAR_WIDTH, (cx->nY+i) * CHAR_HEIGHT, text[i], clrWhite, clrBlack, font);
		}
		
		osd->Flush();
	}
	else
		dprint(1, "No OSD available");
	return true;
}

void CcProcessor::ClearLine(byte line, byte pos, int len){
	cMutexLock lock(&osdMutex);
	if(osd){
		int length = (len < 0 ? OSDWIDTH - pos * font->Width('W') : len * font->Width('W'));	
		osd->DrawRectangle(pos * font->Width('W'),
			font->Height()*line, pos * font->Width('W') + length, font->Height() * (line+1),
			clrTransparent);
		osd->Flush();
	}
}

void CcProcessor::SetRollUp(sCcContext *context, int nRollupMode){
	nRollup = nRollupMode;
	scroller.SetWindow(32,nRollupMode);
	context->nX = 0;
	context->nY = CC_HEIGHT - nRollupMode;
}

void CcProcessor::ClearBuffer(void){
	for(int i=0; i<bufferContext->nBufferPos; i++)				
		if(bufferContext->commands[i].data!=NULL && (bufferContext->commands[i].length>0)){
			//dprint(1, "\nCleaning data...");	
			if(bufferContext->commands[i].data!=NULL)
				delete [] bufferContext->commands[i].data;

			bufferContext->commands[i].data = NULL;
			bufferContext->commands[i].length = 0;
		}	
	bufferContext->nBufferPos = 0;
}
