#ifndef CcProcessor_H_
#define CcProcessor_H_

#include <vdr/osd.h>
#include <vdr/thread.h>
#include "CcExtractor.h"
#include "scroller.h"


#define CHAR_WIDTH font->Width('W')
#define CHAR_HEIGHT font->Height()

#define CC_SCREEN_WIDTH 32
#define CC_SCREEN_HEIGHT 16
//NTSC screen size
#define OSDWIDTH    480 //600 // in pixels
#define OSDHEIGHT   390 //420 // in pixels

#define START_X		120
#define START_Y		45

#define UNDEFINED 0
#define POPUP	1
#define ROLLUP2	2
#define ROLLUP3	3
#define ROLLUP4 4
#define PAINTON	5
#define TEXT	6

#define CC_TEXT 0
#define CC_INFO 1

#define VIRTUAL_BUFFER_SIZE 50

#define NOT_ROLLUP(x) ((x!=ROLLUP2) && (x!=ROLLUP3) && (x!=ROLLUP4))

#define VIRTUAL_SCREEN_SIZE 500

#define isPopup() (style == POPUP)
#define resetBuffer() vsPos = 0; currentContext = &virtualContext[0]
#define isVirtual() (currentContext != &realContext)

typedef struct CcText{
	int posX;
	int posY;
	byte text[CC_SCREEN_WIDTH + 1];
	int length;
	int color;
	bool underline;
	int style;
} CcText;

typedef struct {
	byte color;
	bool bItalic;	
	bool bUnderline;	
	
	int nX, nY; //character level position		
	CC_Data commands[CC_BUFFER_SIZE];
	int nBufferPos;
} sCcContext;

class CcProcessor : public cThread{
public:
	CcProcessor(CcExtractor *x);
	virtual ~CcProcessor();
	void Stop();
protected:
	virtual void Action();
	
	//Process the CC data and returns TRUE when they have to be destroyed
	//bool Process(CC_Data* data);
	
private:
	//tools
	void	ResetBuffer(CcText **buffer, int *size);
	void	PrintBuffer(CcText **buffer, int size);
	bool	CheckPAC(CcText *info, byte hi, byte lo);
	void	Print(CcText *data, bool flush);
	void	ClearScreen(void);
	int		UndoubleCommands(byte* commands, int size, byte *buffer);
	//bool	DoCommand(CcText* info, byte hi, byte lo);
	
	
	void InitOsd();
	void InitContext(CcText *cx);
	
	////
	bool Process(CC_Data* data, bool bWorkInBuffer);
	bool	DoCommand(sCcContext *context, byte hi, byte lo);
	bool WriteInBuffer(CC_Data *data, int pos, int len);
	bool PrintLine(sCcContext *cx, char *row, int len);
	bool PrintRollUp(sCcContext *cx, char **text, int lines);
	void SetRollUp(sCcContext *context, int nRollupMode);
	void ClearLine(byte line, byte pos, int len);
	void ClearBuffer(void);
	////
			
	int vsPos;
		
	
	//the non-empty border
	int firstLine; // = 0;
	int firstColumn;
	int lastLine;
	int lastColumn;
	
	cOsd *osd; // = NULL;
	bool osdInitialized;
	
	const cFont *font;
	
	//the cc extractor
	CcExtractor *extractor;
	
	CcText *currentContext;
	CcText realContext;
	CC_Data *virtualContext[VIRTUAL_BUFFER_SIZE];

	// Style: see above
	int style;
	// Type: 0x00 - CC1 .. 0x03 - CC4 | 0x10 - T1 .. 0x13 - T4 | 0x80 - XDS
	int type; 
	
	//protecting the osd acces
	cMutex osdMutex;
	
	bool bWriteInBuffer;
	sCcContext *crtContext, *bufferContext;
	bool bSkipBuffer;	
	CcScroller scroller;
	int nRollup;
	bool bTextMode;

};

#endif /*CcProcessor_H_*/
