#include "CcStatus.h"
#include "commons.h"

CcStatus::CcStatus()
{
	extractor	= NULL;
	receiver	= NULL;
	isActive = true; //set initially as active even that it will be activated at the first ChannelSwitch call
									//if disabled it will prevent the ChannelSwitch to operate, so it will never get activated
	
	osd			= new CcOsd(this);
}
void CcStatus::CleanUp(){
	dprint(2, "CcStatus::CleanUp()");	  	
	if(extractor!=NULL){		
		dprint(2, "Deleting the extractor...");	  	
		delete extractor;
		extractor = NULL;
	}
	
	if(receiver!=NULL){
		dprint(2, "Deleting the receiver...");	  		
		delete receiver;
		receiver = NULL;
	}

}

CcStatus::~CcStatus()
{
	Stop();
	//start cleaning
	CleanUp();
	
	//shutting down the OSD
	if(osd!=NULL){
		dprint(2, "Deleting the OSD...");	 
		delete osd;
		osd = NULL;
	}	
}

void CcStatus::ChannelSwitch(const cDevice * device, int channelNumber)
{	  
	dprint(2, "CcStatus::ChannelSwitch(channnel::%d)", channelNumber);
	 
	if(!isActive){
		dprint(2, "CcStatus::ChannelSwitch - Status is not active; no reason to restart the status for chn#%d", channelNumber);	  	
		return;
	}
	Stop();	 
  if(channelNumber>0){  	
	 	if(config.getCcActive())	
		  	Start();
	  else
			dprint(2, "CcStatus::ChannelSwitch - Plugin is not active; no reason to start the status for chn#%d", channelNumber);	  	
  }
  else
  	dprint(2, "CcStatus::ChannelSwitch - Invalid channel; no reason to start the status for chn#%d", channelNumber);	  	
}

CcExtractor *CcStatus::GetExtractor(){
	return extractor;
}


void CcStatus::Start(void){
	dprint(1, "CcStatus::Start - Starting CcStatus...");
	if(!isActive){
		int apid[2] = {0, 0};
		int dpid[2] = {0, 0};
		
		CleanUp();

		if(config.getCcActive()){
			dprint(1, "CcStatus::Start - Starting receiver...");
			
			extractor = new CcExtractor(READ_BUFFER);
			extractor->SetChannel(config.getCcType());
					
			cChannel *channel = Channels.GetByNumber(cDevice::CurrentChannel());
			dprint(1, "CcStatus::Start - Starting receiver - channel info retrieved");
			apid[0] = channel->Apid(0);
			dpid[0] = channel->Dpid(0);		
			
			receiver = new CcReceiver(channel->Ca(), channel->Vpid(), apid, dpid);
			dprint(1, "CcStatus::Start - Starting receiver - new receiver created");
			receiver->setExtractor(extractor);
			
			cDevice::ActualDevice()->AttachReceiver(receiver);
			dprint(1, "CcStatus::Start - Starting receiver - receiver attached");		
				
			isActive = true;
		}
		else 
			dprint(1, "CcStatus::Start - Plugin is not active; no reason to start the status.");
	}
	else
		dprint(1, "CcStatus::Start - Status is active; no reason to start it.");
	dprint(1, "CcStatus::Start - finished.");

}

void CcStatus::Stop(void){
	//stop the activity
	dprint(1, "CcStatus::Stop(void)");
	if(isActive){
		if(receiver!=NULL){  	
			dprint(2, "CcStatus::Stop(void) - Detaching the receiver...");	
			cDevice::ActualDevice()->Detach(receiver);
		}
		else
			dprint(1, "CcStatus::Stop(void) - Receiver does not exist!");
		
		//CleanUp();
		isActive = false;
	}
	else
		dprint(1, "CcStatus::Stop(void) - Status is not active; no reason to stop it.");
	dprint(1, "CcStatus::Stop - finished.");

}
